import * as express from 'express';
import * as cookieParser from 'cookie-parser';
import * as bodyParser from 'body-parser';
import { MongoClient, MongoError } from 'mongodb';

import { IOptions } from '../interface/options';
import { IReturn } from '../interface/return';
import { IRoute } from '../interface/route';
import { IContext } from '../interface/context';

export class Server {

	protected server: any;
	protected controllers: Array<any> = [];
	protected models: Array<any> = [];
	protected availableModels: any = {};
	protected client: any;
	protected database: any;

	constructor(protected options: IOptions) {
		this.server = express();
		this.server.use(cookieParser());
		this.server.use(bodyParser.urlencoded({ extended: true }));
		this.server.use(bodyParser.json());
	}

	public async init(): Promise<any> {

		// Prepare and initialise database.
		if (this.options.database && this.options.db_port && this.options.hostname) {
			if (this.options.username && this.options.password) {
				await this.setupDatabase(true);
			} else {
				await this.setupDatabase(false);
			}
		}
	}

	public async setupDatabase(shouldAuth = false): Promise<void> {
		return new Promise((resolve: any, reject: any) => {
			const context = this;
			let database_uri: string;
			if (shouldAuth) {
				database_uri = `mongodb://${this.options.username}:${this.options.password}@${this.options.hostname}:${this.options.db_port}/${this.options.database}`;
			} else {
				database_uri = `mongodb://${this.options.hostname}:${this.options.db_port}/${this.options.database}`;
			}

			MongoClient.connect(database_uri, {
				useUnifiedTopology: true,
				useNewUrlParser: true,
			}, (err: MongoError, client: MongoClient) => {
				if (err) reject(`Failed to connect to database, code: ${err.code} and message: ${err.message}.`);
				context.client = client;
				resolve(true);
			});
		});
	}

	public async start(): Promise<void> {

		// We need to process the models.
		await this.processModels();

		// Then we need to process controllers.
		await this.processControllers();

		this.server.listen(this.options.port, () => {
			console.log('Aether Server, listening on port: ' + this.options.port);
		});
	}

	private async processModels(): Promise<void> {
		for (const index in this.models) {
			const definition: any = Reflect.getMetadata('model', this.models[index]);
			this.database = this.client.db(this.options.database);
			this.availableModels[definition.collection] = new this.models[index](definition, this.database.collection(definition.collection));
		}
	}

	private async processControllers(): Promise<void> {
		this.controllers.forEach((controller: any) => {

			// Get the controller information and meta.
			const instance: any = new controller();
			const prefix: string = Reflect.getMetadata('prefix', controller);
			const routes: Array<IRoute> = Reflect.getMetadata('routes', controller);
			const models: Array<any> = Reflect.getMetadata('models', controller);

			// Now instantiate all required models.
			for (const index in models) {
				if (typeof this.availableModels[models[index]] !== 'undefined') {
					instance.models[models[index]] = this.availableModels[models[index]];
				}
			}

			// Loop the routes and setup the events.
			routes.forEach((route: any) => {
				this.server[route.requestMethod](prefix + route.path, async (request: express.Request, response: express.Response): Promise<void> => {

					// Define headers
					const headers: any = {};
					for (const index in request.headers) {
						headers[index] = request.headers[index];
					}

					// Define query
					const query: any = {};
					for (const index in request.query) {
						query[index] = request.query[index];
					}

					// Define params
					const params: any = {};
					for (const index in request.params) {
						params[index] = request.params[index];
					}

					// Define cookies
					const cookies: any = {};
					for (const index in request.cookies) {
						cookies[index] = request.cookies[index];
					}

					// Create the context object.
					const context: IContext = {
						httpVersion: request.httpVersion,
						headers: headers,
						requestedUrl: request.url,
						httpMethod: request.method,
						ipAddress: request.socket.remoteAddress || '',
						params: params,
						query: query,
						cookies: cookies,
						body: request.body,
						setCookie: (name: string, value: string, maxAge?: number, httpOnly?: boolean, secure?: boolean) => {
							const cookieParams: any = {};
							if (maxAge) cookieParams.maxAge = (1000 * 86400 * maxAge);
							if (httpOnly) cookieParams.httpOnly = httpOnly;
							if (secure) cookieParams.secure = secure;
							response.cookie(name, value, cookieParams);
							return true;
						}
					}

					// Call the function in the callback with the context.
					const result: IReturn = await instance[route.methodName](context);
					response.status(result.status).send(result.content);
				});
			});
		});
	}

	public registerControllers(controllers: Array<any>): void {
		for (const index in controllers) {
			this.controllers.push(controllers[index]);
		}
	}

	public registerModels(models: Array<any>): void {
		for (const index in models) {
			this.models.push(models[index]);
		}
	}

	public registerMiddleware(middleware: (request: express.Request, response: express.Response, next: express.NextFunction) => void): void {
		this.server.use(middleware);
	}
}